/*-
 * Copyright 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Authors: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *	    Nils Goroll <nils.goroll@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Implementation inspired by Varnish cache_esi_deliver.c
 */

/* for SSIZE_MAX */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>

#include "cache/cache_varnishd.h"
#include "cache/cache_transport.h"
#include "cache/cache_filter.h"
#include "cache/cache_vgz.h"
#include "cache/cache_esi.h"
#include "cache/cache_objhead.h"
#include "cache/cache_pool.h"
#include "foreign/from_cache_esi_deliver.h"
#include "storage/storage.h"
#include "storage/storage_simple.h"
#include "vcl.h"
#include "vtim.h"
#include "vend.h"
#include "vgz.h"

#include "foreign/qdef.h"
#include "vqueue.h"

#include "vcc_if.h"

#define VFAIL(ctx, fmt, ...) \
        VRT_fail((ctx), "vdp pesi failure: " fmt, __VA_ARGS__)

/*
 * whether to use blocking threads for final (private/hfm/fpm) objects
 *
 * if false, buffering will be used
 */
static int block_final = 0;
/*
 * whether to push bytes early
 *
 */
static int front_push = 0;

#ifdef DEBUG
#define VSLdbgv(req, fmt, ...)						\
	do {								\
		VSL(SLT_Debug, 0, "xid=%u t=%.6f " fmt,			\
		    VXID(req->vsl->wid), VTIM_real(), __VA_ARGS__);	\
		VSLb(req->vsl, SLT_Debug, fmt, __VA_ARGS__);		\
	} while(0)

#define VSLdbg(req, msg)				\
	do {						\
		VSL(SLT_Debug, 0, "xid=%u t=%.6f " msg,	\
		    VXID(req->vsl->wid), VTIM_real());	\
		VSLb(req->vsl, SLT_Debug, msg);		\
	} while(0)
#define VSL0dbg(...) VSL(SLT_Debug, 0, __VA_ARGS__)
#else
#define VSLdbgv(req, fmt, ...)	(void)0
#define VSLdbg(req, msg)	(void)0
#define VSL0dbg(...)	(void)0
#endif

static const uint8_t gzip_hdr[] = {
	0x1f, 0x8b, 0x08,
	0x00, 0x00, 0x00, 0x00,
	0x00,
	0x02, 0x03
};

#define GZIP_TAILBUF_SZ 13

#define OC_F_FINAL (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP)

enum n_type {
	T_INVALID = 0,
	T_NEXUS,	// can change into T_SUBREQ / T_FINAL / T_DATA
	T_DATA,
	T_CRC,
	T_SUBREQ,
	T_FINAL	// non-ESI pass / hfm / hfp
};

/*
 * see state.dot:
 *
 * ST_DATA:    may never have any children
 *
 *	       T_DATA / T_CRC / T_SUBREQ / T_FINAL
 *
 * ST_PRIVATE: may receive pushes creating children
 *             unpending must not yet touch it
 *             children can be created unlocked
 *             can change into other types
 *
 *             T_NEXUS only
 *
 * ST_OPEN:    may receive pushes creating children
 *             owning thread may run front delivery
 *             any changes below locked only
 *
 *             T_NEXUS only
 *
 *
 * ST_CLOSED:  the request is done, no pushes can occur,
 *             unpending can proceed upwards
 *
 *             T_NEXUS only
 *
 * ST_UNPENDING: in the process of being pushed to the client
 *
 *	         T_DATA / T_CRC / T_SUBREQ / T_FINAL
 *
 * ST_DELIVERED: We have pushed data up
 *
 *               any type
 *
 *		 for T_NEXUS, means "anything below is delivered"
 */
enum n_state {
	ST_INVALID = 0,
	ST_DATA,
	ST_PRIVATE,
	ST_OPEN,
	ST_CLOSED,
	ST_UNPENDING,
	ST_DELIVERED,
};

VSTAILQ_HEAD(node_head, node);

struct node_nexus {
	struct node_head	children;
	struct objcore		*oc;
	struct req		*req;
	const struct worker	*owner; // ST_OPEN only

	/* number of nodes pending under this node while state == ST_PRIVATE */
	int			npending_private;
	/*
	 * crc for data nodes immediately below
	 * updated by push_crc, pretendgzip and gzgz
	 */
	struct nexus_gzip	gzip;	// in foreign
};

enum t_crc {
	INVALID = 0,
	UPDATE,
	FINAL	// combine with parent or gen tail
};

struct node_crc {
	enum t_crc	ctype;
	uint32_t	icrc;
	ssize_t		l_icrc;
};

struct node_data {
	const void		*ptr;
	struct storage		*st;
	ssize_t			len;
	enum vdp_action		act;
};

/*
 * we transfer the sub-request, boc and oc to the topreq thread, delivery
 * happens there
 */
struct node_subreq {
	struct req		*req;
	struct boc		*boc;
	// oc is NULL if already transferred back into req
	struct objcore		*oc;
	// subreq to topreq delivery
	int			done;
	pthread_cond_t		cond;
};

// sub-state for node_final while in ST_DATA
enum fi_state {
	FI_READY	= 0,
	FI_GO,			// topreq signalling req to deliver
	FI_DONE,		// req signalling topreq it is done
	FI_DESTROYED		// cond/mtx destroyed (fini_final())
};

/* we block the sub-thread when it's ready for delivery and continue when the
 * topreqp tells it to */
struct node_final {
	enum fi_state		fi_state;
	pthread_mutex_t	fi_mtx;
	pthread_cond_t		fi_cond;
};

struct node {
	unsigned		magic;
#define NODE_MAGIC 0xe31edef3
	enum n_type		type;
	enum n_state		state;
	VSTAILQ_ENTRY(node)	sibling;
	VSTAILQ_ENTRY(node)	unpend;
	struct node		*parent;

	union {
		struct node_nexus	nexus;	// T_NEXUS
		struct node_data	data;	// T_DATA
		struct node_subreq	subreq; // T_SUBREQ
		struct node_final	final;	// T_FINAL
		struct node_crc	crc;	// T_CRC
	};
};

struct bytes_tree {
	unsigned		magic;
#define BYTES_TREE_MAGIC 0x49c59d46
	struct lock		tree_lock;
	// esi_level > 1 signalling new data
	pthread_cond_t		cond;
	struct node		*root;
	struct node		*front;
	const struct worker	*front_owner;
	const struct worker	*unpend_owner;
	int			npending;
	int			retval;
};

/*
 * pesi_tree is a global context for the entire ESI tree started at level 0.
 * pesi is local to each subrequest, allocated in req WS.
 *
 * XXX merge pesi_tree with bytes_tree ?
 */

struct pesi;

struct pesi_tree {
	unsigned		magic;
#define PESI_TREE_MAGIC 0xe8ce8adb
	struct bytes_tree	*tree;

	/* all pesis (struct pesi) we started */
	struct lock		task_lock;
	pthread_cond_t		task_cond;
	VTAILQ_HEAD(,pesi)	task_head;
	int			task_running;
	int			task_finishing;
};

#define PF_HAS_TASK		1U
/* vcl-controlled flags */
#define PF_CFG_SERIAL		(1U<<1)
#define PF_CFG_THREAD		(1U<<2)
/* undocumented for now */
#define PF_CFG_BLOCK_FINAL	(1U<<3)
#define PF_CFG_FRONT_PUSH	(1U<<4)

#define PF_CFG_DEFAULT				\
	( PF_CFG_THREAD			\
	| block_final ? PF_CFG_BLOCK_FINAL : 0	\
	| front_push ? PF_CFG_FRONT_PUSH : 0	\
	)

#define PF_MASK_CFG		\
	( PF_CFG_SERIAL	\
	| PF_CFG_THREAD	\
	| PF_CFG_BLOCK_PRIVATE	\
	| PF_CFG_BLOCK_FINAL	\
	)
/*
 * per request state
 *
 * pecx: ESI parser
 * pesi: any request
 *
 */
struct pecx {
	unsigned		magic;
#define PECX_MAGIC		0x5d8cd06d
	const uint8_t		*p;
	const uint8_t		*e;

	ssize_t			l;
	int			state;
};

struct pesi {
	unsigned		magic;
#define PESI_MAGIC 0xa6ba54a0
	unsigned		flags;

	struct pesi_tree	*pesi_tree;
	struct worker		*wrk;
	struct node		*node;
	int			woken;

	struct pecx		pecx[1];
	VTAILQ_ENTRY(pesi)	list;
};

/* Forward declarations */
const struct vdp VDP_pesi_buf;
const struct vdp VDP_pesi;
static const struct transport VPED_transport;
static void req_fini(struct req **, struct worker *);
static void vped_close_vdp(struct req *, int, const struct vdp *);
static void assert_vdp_next_not(struct req *, const struct vdp *);

// node finalizers
static void fini_final(struct req *, struct node *);
static void fini_subreq(struct req *, struct node *);
static void fini_data(struct req *, struct node *);

/* id for PRIV_TASK */
const void * const priv_task_id_serial = &priv_task_id_serial;

/* shared object globals */
static unsigned loadcnt = 0, warmcnt = 0;
static struct VSC_lck *lck_bytes_tree, *lck_pesi_tree;
static struct vsc_seg  *vsc_seg = NULL;
static struct mempool *mempool = NULL;
static unsigned node_alloc_sz;

/* also used by our version of cache_esi_deliver.c */

/* ------------------------------------------------------------
 * basically ved_ved / ved_bytes without the ecx
 * (may be an intermediate solution)
 */

static int v_matchproto_(vdp_fini_f)
vped_to_parent_fini(struct req *req, void **priv)
{
	(void)req;
	*priv = NULL;
	return (0);
}

static int v_matchproto_(vdp_bytes_f)
vped_to_parent_bytes(struct req *req, enum vdp_action act, void **priv,
    const void *ptr, ssize_t len)
{
	struct req *preq;

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CAST_OBJ_NOTNULL(preq, *priv, REQ_MAGIC);
	req->acct.resp_bodybytes += len;
	return (VDP_bytes(preq, act, ptr, len));
}

static const struct vdp vped_to_parent = {
	.name =         "V2P",
	.bytes =        vped_to_parent_bytes,
	.fini =         vped_to_parent_fini,
};



/* ------------------------------------------------------------
 * pesi tasks
 *
 * in the top request, we need to wait for all pesis to finish because, via
 * req->topreq and req->sp, they all reference the top request
 *
 * for this purpose, simple refcounting would suffice, but putting all the pesis
 * on a list does not add any relevant amount of complexity and might help with
 * troubleshooting and through additional assertions
 *
 * pesi_new always happens in a task but the destruction can happen in any
 * order:
 * - for T_SUBREQ, pesi outlives the task
 * - otherwise, pesi gets destroyed before the task
 */
static struct pesi *
pesi_new(struct ws *ws, struct pesi_tree *pesi_tree)
{
	struct pesi *pesi;

	pesi = WS_Alloc(ws, sizeof *pesi);
	if (pesi == NULL)
		return (NULL);

	INIT_OBJ(pesi, PESI_MAGIC);
	pesi->pecx->magic = PECX_MAGIC;
	pesi->pesi_tree = pesi_tree;
	pesi->flags = PF_HAS_TASK | PF_CFG_DEFAULT;

	Lck_Lock(&pesi_tree->task_lock);
	VTAILQ_INSERT_TAIL(&pesi_tree->task_head, pesi, list);
	assert(pesi_tree->task_running >= 0);
	pesi_tree->task_running++;
	Lck_Unlock(&pesi_tree->task_lock);
	return (pesi);
}

/*
 * shutting down a pesi request is a two stage process, because Req_Cleanup() in
 * req_fini() called from ved_task() zeroes the workspace including our
 * per-request struct pesi, so removing the registration from
 * pesi_tree->task_head need to happen before Req_Cleanup().
 *
 * But notification of the topreq needs to happen after Req_Cleanup(), because
 * it still references topreq and sp, thus the topreq may not finish until all
 * pesi tasks have finished.
 *
 * on the implementation:
 *
 * - in vdp_pesi_fini() we call pesi_destroy() to clean up the per-request
 *   struct pesi and remove it from the global list.
 *
 *   For possible debuggung purposes, we record the fact that a pesi request is
 *   finishing
 *
 * - in ved_task(), we call task_fini() to decrement the global counter and
 *   notify the top request
 *
 * for esi_level == 0, all cleanup happens in vdp_pesi_fini
 */

static void
pesi_destroy(struct pesi **pesip)
{
	struct pesi *pesi;
	struct pesi_tree *pesi_tree;

	TAKE_OBJ_NOTNULL(pesi, pesip, PESI_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pecx, PECX_MAGIC);
	TAKE_OBJ_NOTNULL(pesi_tree, &pesi->pesi_tree, PESI_TREE_MAGIC);

	Lck_Lock(&pesi_tree->task_lock);
	VTAILQ_REMOVE(&pesi_tree->task_head, pesi, list);

	assert(pesi_tree->task_running >= 0);

	if (pesi->flags & PF_HAS_TASK)
		pesi_tree->task_finishing++;
	Lck_Unlock(&pesi_tree->task_lock);

	memset(pesi, 0, sizeof *pesi);
}


static void
task_fini(struct pesi_tree *pesi_tree, struct pesi *pesi)
{
	Lck_Lock(&pesi_tree->task_lock);
	assert(pesi_tree->task_running > 0);
	if (pesi == NULL) {
		assert(pesi_tree->task_finishing > 0);
		pesi_tree->task_finishing--;
	}
	else {
		AN(pesi->flags & PF_HAS_TASK);
		pesi->flags &= ~PF_HAS_TASK;
	}
	pesi_tree->task_running--;
	if (pesi_tree->task_running == 0) {
		AZ(pesi_tree->task_finishing);
		AZ(pthread_cond_signal(&pesi_tree->task_cond));
	}
	Lck_Unlock(&pesi_tree->task_lock);
}

/* ------------------------------------------------------------
 * tree related
 */

static inline struct node *
node_alloc(void)
{
	struct node *node;
	unsigned sz;

	AN(mempool);
	node = MPL_Get(mempool, &sz);
	MPL_AssertSane(node);
	assert(sz >= sizeof *node);
	// MPL zeroes
	AZ(node->magic);
	node->magic = NODE_MAGIC;
	return (node);
}

static void
node_fini(struct req *req, struct node *node)
{
	switch (node->type) {
	case T_FINAL:
		fini_final(req, node);
		break;
	case T_SUBREQ:
		fini_subreq(req, node);
		break;
	case T_DATA:
		fini_data(req, node);
		break;
	case T_CRC:
		break;
	case T_NEXUS:
		/* root req owned by varnish-cache */
		if (node->parent == NULL)
			break;
		req_fini(&node->nexus.req, req->wrk);
		break;
	default:
		INCOMPL();
	}
}

/* recursively fini all nodes */
static inline void
tree_fini(struct req *req, struct node *node)
{
	struct node *child;

	if (node->type == T_NEXUS)
		VSTAILQ_FOREACH(child, &node->nexus.children, sibling)
			tree_fini(req, child);

	node_fini(req, node);
}

/* recursively free all nodes == free the tree */
static void
tree_free(struct req *req, struct node *node)
{
	struct node *child, *tmp;

	AN(mempool);
	MPL_AssertSane(node);

	if (node->type == T_NEXUS)
		VSTAILQ_FOREACH_SAFE(child, &node->nexus.children, sibling, tmp)
			tree_free(req, child);

	node_fini(req, node);
	MPL_Free(mempool, node);
}

static void
node_insert(struct bytes_tree *tree, struct node *parent,
    struct node *node)
{
	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(parent, NODE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	assert(parent->type == T_NEXUS);

	if (parent->state == ST_OPEN)
		Lck_Lock(&tree->tree_lock);
	else
		assert(parent->state == ST_PRIVATE);

	switch (node->type) {
	case T_NEXUS:
		assert(node->state == ST_PRIVATE ||
		       node->state == ST_OPEN);
		assert(node->nexus.gzip.magic == NEXUS_GZIP_MAGIC);
		break;
	case T_CRC:
	case T_DATA:
		assert(node->state == ST_DATA);
		break;
	case T_SUBREQ:
	case T_FINAL:
	default:
		/* cannot insert SUBREQ and FINAL yet */
		INCOMPL();
	}

	AZ(node->parent);

	node->parent = parent;
	if (node->type == T_NEXUS)
		VSTAILQ_INIT(&node->nexus.children);

	VSTAILQ_INSERT_TAIL(&parent->nexus.children, node, sibling);

	assert(tree->npending >= 0);

	if (parent->state == ST_PRIVATE) {
		parent->nexus.npending_private++;
	}
	else {
		assert(parent->state == ST_OPEN);
		AZ(parent->nexus.npending_private);
		tree->npending++;

		if (node->state == ST_DATA)
			AZ(pthread_cond_signal(&tree->cond));

		Lck_Unlock(&tree->tree_lock);
	}

	VSLdbgv(tree->root->nexus.req, "node_insert: node=%p parent=%p "
	    "parent->state=%d parent->npending_private=%d "
	    "npending=%d",
	    node, parent,
	    parent->state, parent->nexus.npending_private,
	    tree->npending);
}

/*
 * node mutation: turn an existing nexus into something else
 *
 * howto:
 *
 * 0) node must be T_NEXUS / ST_PRIVATE, untouched
 *
 * 1) call node_mutate_prep()
 *
 * 2) modify data in the destination-type specific struct
 *
 * 3) call node_mutate_lock() with destination type/state
 *
 * 4) optionally do more things under the lock
 *
 * 5) call node_mutate_unlock()
 *
 */
static void
node_mutate_prep(struct bytes_tree *tree, struct node *node)
{
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	CHECK_OBJ_NOTNULL(node->parent, NODE_MAGIC);

	assert(node->state == ST_PRIVATE);
	assert(node->type == T_NEXUS);
	assert(node != tree->root);

	AZ(node->nexus.npending_private);
	AZ(node->nexus.oc);

	memset(&node->nexus, 0, sizeof node->nexus);
}

static inline void
node_mutate_lock(struct bytes_tree *tree, struct node *node,
		 enum n_type type, enum n_state state)
{
	/* these checks can be relexed when needed */
	assert(type == T_DATA ||
	       type == T_SUBREQ ||
	       type == T_FINAL);
	assert(state == ST_DATA);

	Lck_Lock(&tree->tree_lock);
	node->type = type;
	node->state = state;
	if (node->parent->state != ST_PRIVATE)
		AZ(pthread_cond_signal(&tree->cond));
}

static inline void
node_mutate_unlock(struct bytes_tree *tree)
{
	Lck_Unlock(&tree->tree_lock);
}

/*
 * CHK_ORDER: within child list can only transition from ST_DELIVERED -> !
 * ST_DELIVERED iow, once a child is found pending, all others must be also
 */

enum check_state {
	CHK_ANY = 0,
	CHK_PEND,	// ! ST_DELIVERED && ! ST_UNPENDING
	CHK_DELI,	// ST_DELIVERED
	CHK_ORDER
};

/*
 * keep in prod builds for now to gain trust in tree sanity
 *
 * XXX enable only for DEBUG later?
 */
#ifdef NO_ASSERT_NODE
#define assert_node(n, c) (void)0
#define assert_nexus(n, c) (void)0
#else
static void assert_node(struct node *node, enum check_state check);

static void
assert_nexus(struct node *node, enum check_state check)
{
	struct node *child;

	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	assert(node->type == T_NEXUS);

	VSTAILQ_FOREACH(child, &node->nexus.children, sibling) {
		assert_node(child, check);
		if (check == CHK_ORDER &&
		    child->state != ST_DELIVERED &&
		    child->state != ST_UNPENDING)
			check = CHK_PEND;
	}
}

static void
assert_node(struct node *node, enum check_state check)
{
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	switch (check) {
	case CHK_ANY:
	case CHK_ORDER:
		break;
	case CHK_PEND:
		assert(node->state != ST_DELIVERED &&
		       node->state != ST_UNPENDING);
		break;
	case CHK_DELI:
		assert(node->state == ST_DELIVERED);
		break;
	default:
		INCOMPL();
	}

	switch (node->state) {
	case ST_PRIVATE:
		assert(node->type == T_NEXUS);
		/* hands off all nexus fields */
		break;
	case ST_DATA:
	case ST_UNPENDING:
		assert(node->type != T_NEXUS);
		break;
	case ST_OPEN:
		AN(node->nexus.owner);
		AZ(node->nexus.npending_private);
		assert_nexus(node, CHK_ORDER);
		break;
	case ST_CLOSED:
		AZ(node->nexus.owner);
		AZ(node->nexus.npending_private);
		assert_nexus(node, CHK_ORDER);
		break;
	case ST_DELIVERED:
		if (node->type == T_NEXUS) {
			assert_nexus(node, CHK_DELI);
			AZ(node->nexus.oc);
		} else if (node->type == T_DATA) {
			AZ(node->data.st);
		}
		break;
	default:
		INCOMPL();
	}
}
#endif

static void
set_unpending(struct bytes_tree *tree, struct node *node)
{
	struct node *parent;

	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	parent = node->parent;
	CHECK_OBJ_ORNULL(parent, NODE_MAGIC);

	Lck_AssertHeld(&tree->tree_lock);

	assert(node->state == ST_DATA);
	assert(node->type == T_DATA ||
	       node->type == T_SUBREQ ||
	       node->type == T_FINAL ||
	       node->type == T_CRC);

	node->state = ST_UNPENDING;

	assert_node(node, CHK_ANY);

	if (parent == NULL)
		return;

	assert(parent->state == ST_OPEN || parent->state == ST_CLOSED);

}

static void
set_delivered(struct bytes_tree *tree, struct node *node)
{
	struct node *parent;
	struct req *req;

	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	parent = node->parent;
	CHECK_OBJ_ORNULL(parent, NODE_MAGIC);

	Lck_AssertHeld(&tree->tree_lock);

	assert(node->state == ST_DATA ||
	       node->state == ST_CLOSED ||
	       node->state == ST_UNPENDING);

	if (node->type == T_NEXUS &&
	    node->nexus.oc != NULL) {
		req = tree->root->nexus.req;
		CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
		CHECK_OBJ_NOTNULL(req->wrk, WORKER_MAGIC);

		assert(node->state == ST_CLOSED);
		(void) HSH_DerefObjCore(req->wrk, &node->nexus.oc, 0);
	}

	node->state = ST_DELIVERED;
	assert(tree->npending > 0);
	tree->npending--;

	assert_node(node, CHK_DELI);

	if (parent == NULL) {
		/* Response has esi:remove but no esi:include. */
		AN(node);
		assert(node == tree->root);
		VSL0dbg("set_delivered: root node=%p", node);
		return;
	}

	assert(parent->state == ST_OPEN || parent->state == ST_CLOSED);

	if (parent->state == ST_OPEN || VSTAILQ_NEXT(node, sibling) != NULL) {
		VSL0dbg("set_delivered: stop parent=%p state=%d "
		    "sibling %p", parent, parent->state,
		    VSTAILQ_NEXT(node, sibling));
		return;
	}

	set_delivered(tree, parent);
}

/* ST_PRIVATE -> ST_OPEN transition
 *
 * child nodes may get unpended by owner
 * any access locked only
 */

static void
set_open(struct bytes_tree *tree, struct node *node, const struct worker *wrk)
{
	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	Lck_AssertHeld(&tree->tree_lock);

	VSL0dbg("set_open node=%p wrk=%p", node, wrk);

	assert(node->state == ST_PRIVATE);
	AZ(node->nexus.owner);

	node->state = ST_OPEN;
	tree->npending += node->nexus.npending_private;
	node->nexus.npending_private = 0;
	node->nexus.owner = wrk;
}

/* (ST_PRIVATE | ST_OPEN) -> CLOSE transition
 *
 * possible states and transitions
 *
 * undelivered children: -> ST_CLOSED
 * all children delivered: -> ST_DELIVERED
 */

static void
set_closed(struct bytes_tree *tree, struct node *node, const struct worker *wrk)
{
	struct node *child;

	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	Lck_AssertHeld(&tree->tree_lock);

	if (node->state == ST_PRIVATE)
		set_open(tree, node, wrk);

	VSL0dbg("set_closed node=%p wrk=%p", node, wrk);

	assert(node->state == ST_OPEN);
	assert(node->nexus.owner == wrk);
	assert_node(node, CHK_ORDER);

	node->state = ST_CLOSED;
	node->nexus.owner = NULL;

	/* unpending may wait for this node to close */
	AZ(pthread_cond_signal(&tree->cond));

	if (node->type == T_NEXUS)
		VSTAILQ_FOREACH(child, &node->nexus.children, sibling)
			if (child->state != ST_DELIVERED)
				return;

	set_delivered(tree, node);
}

/*--------------------------------------------------------------------*/

static void  v_matchproto_(task_func_t)
ved_task(struct worker *wrk, void *priv)
{
	struct req *req;
	struct sess *sp;
	struct pesi *pesi;
	struct pesi_tree *pesi_tree;
	struct node *node;
	enum req_fsm_nxt s;

	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);

	AZ(wrk->task.func);
	AZ(wrk->task.priv);

	/*
	 * if we got an aws reservation, we have been run via Pool_Task_Arg() or
	 * the direct call, otherwise via Pool_Task()
	 */
	if (wrk->aws->r != NULL) {
		priv = *(void **)priv;
		WS_Release(wrk->aws, 0);
	}
	CAST_OBJ_NOTNULL(req, priv, REQ_MAGIC);

	VSLdbgv(req, "ved_task: req=%p", req);

	assert(req->esi_level > 0);
	sp = req->sp;
	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CAST_OBJ_NOTNULL(pesi, req->transport_priv, PESI_MAGIC);
	pesi_tree = pesi->pesi_tree;
	CHECK_OBJ_NOTNULL(pesi_tree, PESI_TREE_MAGIC);
	pesi->wrk = wrk;
	node = pesi->node;
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	CHECK_OBJ_NOTNULL(pesi_tree->tree, BYTES_TREE_MAGIC);

	THR_SetRequest(req);
	VSLb_ts_req(req, "Start", W_TIM_real(wrk));

#ifdef DEBUG
	VSLdbgv(req, "ved_task: new req->vdc=%p nxt=%p", req->vdc,
		req->vdc->nxt);
	if (VTAILQ_EMPTY(&req->vdc->vdp))
		VSLdbg(req, "ved_task: VDP list empty");
	else {
		struct vdp_entry *vdpe = VTAILQ_LAST(&req->vdc->vdp,
						     vdp_entry_s);
		CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
		VSLdbgv(req, "ved_task: last VDP=%s", vdpe->vdp->name);
	}
#endif

	CNT_Embark(wrk, req);
	VCL_TaskEnter(req->vcl, req->privs);

	while (1) {
		req->wrk = wrk;
		pesi->woken = 0;
		VSLdbgv(req, "Starting CNT_Request() req=%p", req);
		s = CNT_Request(req);
		VSLdbgv(req, "CNT_Request()=%d req=%p", s, req);
		if (s == REQ_FSM_DONE)
			break;
		DSL(DBG_WAITINGLIST, req->vsl->wid,
		    "loop waiting for ESI (%d)", (int)s);
		assert(s == REQ_FSM_DISEMBARK);
		Lck_Lock(&req->sp->mtx);
		if (!pesi->woken) {
			VSLb(req->vsl, SLT_Debug, "Waiting for busyobj");
			(void)Lck_CondWait(
			    &wrk->cond, &req->sp->mtx, 0);
			VSLb(req->vsl, SLT_Debug, "busyobj wakeup");
		}
		Lck_Unlock(&req->sp->mtx);
		pesi->woken = 0;
		AZ(req->wrk);
		VSLdbg(req, "Calling CNT_Embark()");
		CNT_Embark(wrk, req);
	}

	VCL_Rel(&req->vcl);

	switch (node->type) {
	case T_FINAL:
		assert(pesi == req->transport_priv);
		req->transport_priv = NULL;
		pesi_destroy(&pesi);
		req_fini(&req, wrk);
		AZ(pesi);
		task_fini(pesi_tree, pesi);
		break;
	case T_SUBREQ:
		assert (node->type == T_SUBREQ);
		assert(pesi == req->transport_priv);
		task_fini(pesi_tree, pesi);
		Lck_Lock(&pesi_tree->tree->tree_lock);
		node->subreq.done = 1;
		AZ(pthread_cond_signal(&node->subreq.cond));
		Lck_Unlock(&pesi_tree->tree->tree_lock);
		break;
	case T_NEXUS:
		assert(node->nexus.req == req);
		assert(node->state == ST_OPEN ||
		       node->state == ST_PRIVATE);

		if (VSTAILQ_FIRST(&node->nexus.children) != NULL) {
			/*
			 * normal case: the nexus has received some children
			 *
			 * pesi was destroyed within the FSM via vdp_pesi_fini
			 */
			AZ(req->transport_priv);
			pesi = NULL;

			Lck_Lock(&pesi_tree->tree->tree_lock);
			set_closed(pesi_tree->tree, node, wrk);
			Lck_Unlock(&pesi_tree->tree->tree_lock);
			AZ(pesi);
			task_fini(pesi_tree, pesi);
			break;
		}

		/*
		 * we got nothing for this T_NEXUS:
		 *
		 * - get rid of the request
		 * - turn into a zero data node
		 *
		 * the req may or may not have been fini't in FSM, depending on
		 * error conditions. So we sanity check transport_priv and
		 * destroy pesi if needed
		 */

		// finish the request
		VDP_close(req);

		if (req->transport_priv != NULL)
			assert(req->transport_priv == pesi);

		pesi = req->transport_priv;
		req->transport_priv = NULL;

		if (pesi)
			pesi_destroy(&pesi);

		assert(req == node->nexus.req);
		req = NULL;

		req_fini(&node->nexus.req, wrk);

		/* turn into T_DATA node */
		node_mutate_prep(pesi_tree->tree, node);
		memset(&node->data, 0, sizeof node->data);

		node_mutate_lock(pesi_tree->tree, node,
		    T_DATA, ST_DATA);
		node_mutate_unlock(pesi_tree->tree);

		task_fini(pesi_tree, pesi);
		break;
	default:
		INCOMPL();
	}
}

/*--------------------------------------------------------------------*/

static void
req_fini(struct req **reqp, struct worker *wrk)
{
	struct req *req;

	AN(reqp);
	if (*reqp == NULL)
		return;

	TAKE_OBJ_NOTNULL(req, reqp, REQ_MAGIC);

	VSLdbg(req, "req_fini called");

	VDP_close(req);
	Req_Cleanup(req->sp, wrk, req);
	Req_Release(req);
}

static int
want_serial(struct req *preq)
{
	struct vmod_priv *vclserial;
	struct vrt_ctx dummy_ctx[1];

	INIT_OBJ(dummy_ctx, VRT_CTX_MAGIC);
	dummy_ctx->req = preq;
	dummy_ctx->ws = preq->ws;

	vclserial = VRT_priv_task(dummy_ctx, priv_task_id_serial);

	return (vclserial != NULL && vclserial->priv == (void *)PF_CFG_SERIAL);
}

static int
vped_include(struct req *preq, const char *src, const char *host,
	     struct pesi *pesi, struct node *node)
{
	struct worker *wrk;
	struct sess *sp;
	struct pesi_tree *pesi_tree;
	struct req *req;
	struct bytes_tree *tree;
	struct pesi *pesi2;

	VSLdbg(preq, "ved_include: enter");
	CHECK_OBJ_NOTNULL(preq, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(preq->topreq, REQ_MAGIC);
	sp = preq->sp;
	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
	CHECK_OBJ_NOTNULL(pesi, PESI_MAGIC);
	pesi_tree = pesi->pesi_tree;
	CHECK_OBJ_NOTNULL(pesi_tree, PESI_TREE_MAGIC);
	tree = pesi_tree->tree;
	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	wrk = preq->wrk;

	if ((unsigned)preq->esi_level >= cache_param->max_esi_depth) {
		VSLb(preq->vsl, SLT_VCL_Error,
		    "ESI depth limit reach (param max_esi_depth = %u)",
		    cache_param->max_esi_depth);
		return (-1);
	}

	req = Req_New(wrk, sp);
	AN(req);
	VSLdbgv(preq, "ved_include: new req=%p", req);
	AZ(req->vsl->wid);
	req->vsl->wid = VXID_Get(wrk, VSL_CLIENTMARKER);
	VSLdbgv(preq, "ved_include: new xid=%d", VXID(req->vsl->wid));

	VSLb(req->vsl, SLT_Begin, "req %u esi", VXID(preq->vsl->wid));
	VSLb(preq->vsl, SLT_Link, "req %u esi", VXID(req->vsl->wid));

	req->esi_level = preq->esi_level + 1;

	req->topreq = preq->topreq;

	HTTP_Setup(req->http, req->ws, req->vsl, SLT_ReqMethod);
	HTTP_Dup(req->http, preq->http0);

	http_SetH(req->http, HTTP_HDR_URL, src);
	if (host != NULL && *host != '\0')  {
		http_Unset(req->http, H_Host);
		http_SetHeader(req->http, host);
	}

	http_ForceField(req->http, HTTP_HDR_METHOD, "GET");
	http_ForceField(req->http, HTTP_HDR_PROTO, "HTTP/1.1");

	/* Don't allow conditionals, we can't use a 304 */
	http_Unset(req->http, H_If_Modified_Since);
	http_Unset(req->http, H_If_None_Match);

	/* Don't allow Range */
	http_Unset(req->http, H_Range);

	/* Set Accept-Encoding according to what we want */
	http_Unset(req->http, H_Accept_Encoding);

	CHECK_OBJ_NOTNULL(node->parent, NODE_MAGIC);
	assert(node->parent->type == T_NEXUS);
	if (node->parent->nexus.gzip.is)
		http_ForceHeader(req->http, H_Accept_Encoding, "gzip");

	/* Client content already taken care of */
	http_Unset(req->http, H_Content_Length);
	req->req_body_status = REQ_BODY_NONE;

	AZ(req->vcl);
	if (req->topreq->vcl0)
		req->vcl = req->topreq->vcl0;
	else
		req->vcl = preq->vcl;
	VCL_Ref(req->vcl);

	req->req_step = R_STP_TRANSPORT;
	req->t_req = preq->t_req;

#ifdef DEBUG
	VSLdbgv(preq, "ved_include: preq->vdc=%p nxt=%p", preq->vdc,
		preq->vdc->nxt);
	if (VTAILQ_EMPTY(&preq->vdc->vdp))
		VSLdbg(preq, "ved_include: preq VDP list empty");
	else if (req->esi_level > 1) {
		struct vdp_entry *vdpe = VTAILQ_LAST(&preq->vdc->vdp,
						     vdp_entry_s);
		CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
		VSLdbgv(preq, "ved_include: preq last VDP=%s", vdpe->vdp->name);
	}
#endif

	req->transport = &VPED_transport;

	pesi2 = pesi_new(req->ws, pesi_tree);
	if (pesi2 == NULL) {
		VSLb(req->vsl, SLT_Error,
		     "Cannot allocate workspace for parallel ESI data");
		req_fini(&req, wrk);
		tree->retval = -1;
		return (-1);
	}

	pesi2->node = node;

	req->transport_priv = pesi2;

	req->ws_req = WS_Snapshot(req->ws);

	assert(node->type == T_NEXUS);
	node->nexus.req = req;

	VSLdbgv(preq, "ved_include: attempt new thread req=%p", req);

	if ((pesi->flags & PF_CFG_BLOCK_FINAL) == 0) {
		if (want_serial(preq)) {
			ved_task(wrk, req);
			return (0);
		}

		if (Pool_Task_Arg(wrk, TASK_QUEUE_REQ,
		    ved_task, &req, sizeof req) == 1)
			return (1);
		/*
		 * XXX @geoff add counter
		 *
		 * we can not use the self-rescheduling facility of
		 * Pool_Task_Arg because we cannot unschedule ourself
		 */
		wrk->task.func = NULL;
		wrk->task.priv = NULL;
		WS_Release(wrk->aws, 0);
		ved_task(wrk, req);
		return (0);
	}

	AN(pesi->flags & PF_CFG_BLOCK_FINAL);
	/* XXX because of T_FINAL blocking in the subreq thread,
	 * these threads *have* to run.
	 *
	 * ordering should ensure that we run top->bottom left-right in the ESI
	 * tree, so delivery should be able to let T_FINAL threads run, and
	 * anything but T_FINAL should terminate eventually, but if we spend all
	 * available threads on T_FINAL, we're doomed.
	 *
	 * -- find a way to reserve or check for T_FINAL
	 */
	req->task.func = ved_task;
	req->task.priv = req;
	AZ(Pool_Task(wrk->pool, &req->task, TASK_QUEUE_RUSH));
	return (1);
}

/*--------------------------------------------------------------------*/

#ifdef DEBUG
#include <stdio.h>
#define Debug(fmt, ...) printf(fmt, __VA_ARGS__)
#else
#define Debug(fmt, ...) /**/
#endif

static ssize_t
ved_decode_len(struct req *req, const uint8_t **pp)
{
	const uint8_t *p;
	ssize_t l;

	p = *pp;
	switch (*p & 15) {
	case 1:
		l = p[1];
		p += 2;
		break;
	case 2:
		l = vbe16dec(p + 1);
		p += 3;
		break;
	case 8:
		l = vbe64dec(p + 1);
		p += 9;
		break;
	default:
		VSLb(req->vsl, SLT_Error,
		    "ESI-corruption: Illegal Length %d %d\n", *p, (*p & 15));
		WRONG("ESI-codes: illegal length");
	}
	*pp = p;
	assert(l > 0);
	return (l);
}

/*
 * to minimize lock contention, we walk the tree and gather work on a list
 */

#define set_front(tree, node, wrk) do {					\
		(tree)->front = node;					\
		(tree)->front_owner = wrk;				\
	} while(0)

static void
bytes_unpend_worklist(struct req *req, struct bytes_tree *tree,
    struct node_head *work)
{
	struct node *node, *next;
	const void *p;
	enum check_state check = CHK_DELI;

	(void) check;	// for NODEBUG

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	AN(work);
	assert(VSTAILQ_EMPTY(work));

	Lck_AssertHeld(&tree->tree_lock);
	VSLdbgv(req, "bytes_unpend: retval=%d front=%p root=%p",
		tree->retval, tree->front, tree->root);
	if (tree->retval)
		return;

	CHECK_OBJ_NOTNULL(tree->front, NODE_MAGIC);
	CHECK_OBJ_NOTNULL(tree->root, NODE_MAGIC);

	if (tree->root->state == ST_DELIVERED) {
		VSLdbg(req, "bytes_unpend: whole tree is delivered");
		assert_node(tree->root, CHK_DELI);
		return;
	}

	/*
	 * if unpending changes between threads, we need to make sure that we
	 * only descend into ST_OPEN T_NEXUS which we own, so re-start at the
	 * root
	 *
	 * this might seem inefficient, but we skip over any delivered subtrees,
	 * so we will not walk the whole tree each time
	 */
	AN(tree->front);

	if (tree->front_owner == req->wrk)
		node = tree->front;
	else
		node = tree->root;

	/* when starting, we do not want to see any ST_UNPENDING nodes */
	check = CHK_DELI;

	while (node != NULL) {
		assert_node(node, CHK_ANY);

		while (node->state == ST_DELIVERED ||
		       node->state == ST_UNPENDING) {
			assert_node(node, check);
			VSLdbgv(req, "bytes_unpend: delivered/unpending "
				"node %p state %u", node, node->state);

			next = VSTAILQ_NEXT(node, sibling);
			if (next != NULL) {
				node = next;
				continue;
			}

			if (node->parent && node->parent->state == ST_OPEN) {
				VSLdbgv(req, "bytes_unpend: return at "
					"node %p state %u - await sibling",
					node, node->state);
				set_front(tree, node, req->wrk);
				return;
			}

			/* up and right */
			node = node->parent;
			while (node != NULL) {
				next = VSTAILQ_NEXT(node, sibling);
				if (next != NULL) {
					node = next;
					continue;
				}
				node = node->parent;
			}

			AZ(node);
			VSLdbg(req, "bytes_unpend: reached root");
			set_front(tree, tree->root, req->wrk);
			return;
		}

		if (node->state == ST_PRIVATE) {
			VSLdbgv(req, "bytes_unpend: return at "
				"node %p state %u - private",
				node, node->state);
			set_front(tree, node, req->wrk);
			return;
		}

		assert_node(node, CHK_PEND);

		/* for ST_OPEN, only descend if we own it */
		if (node->type == T_NEXUS) {
			next = VSTAILQ_FIRST(&node->nexus.children);

			switch (node->state) {
			case ST_CLOSED:
				// closed node with no children must not exist
				AN(next);
				node = next;
				continue;
			case ST_OPEN:
				set_front(tree, node, req->wrk);
				AN(node->nexus.owner);
				if (node->nexus.owner == req->wrk) {
					node = next;
					continue;
				}
				return;
			default:
				INCOMPL();
			}
		}

		assert(node->state == ST_DATA);

		if (node->type == T_SUBREQ ||
		    node->type == T_FINAL ||
		    node->type == T_CRC) {
			VSTAILQ_INSERT_TAIL(work, node, unpend);
			set_unpending(tree, node);
			check = CHK_ANY;
			continue;
		}

		assert(node->type == T_DATA);

		p = node->data.ptr;
		if (p == NULL && node->data.len > 0) {
			CHECK_OBJ_NOTNULL(node->data.st, STORAGE_MAGIC);
			p = node->data.st->ptr;
		}

		if (p == NULL || node->data.len == 0) {
			set_delivered(tree, node);
			continue;
		}

		VSTAILQ_INSERT_TAIL(work, node, unpend);
		set_unpending(tree, node);

		// for more iterations, also accept UNPEND
		// for now, we use any
		check = CHK_ANY;
	}
}
#undef set_front

/*
 * restore the original subreq state and take care that we do not mess up flags,
 * objcore etc
*/
static void
subreq_fixup(struct node *node, struct req *req)
{
	struct req *subreq;

	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);
	assert(node->type == T_SUBREQ);
	AN(node->subreq.done);
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	subreq = node->subreq.req;
	CHECK_OBJ_NOTNULL(subreq, REQ_MAGIC);

	/*
	 * we want to make sure there is only ever one reference to the objcore.
	 * as we get called from push_subreq() and fini_subreq(), depending on
	 * if push_subreq() has already been called, the transfer has already
	 * happened or not
	 */

	if (subreq->objcore != NULL) {
		CHECK_OBJ_NOTNULL(subreq->objcore, OBJCORE_MAGIC);
		AZ(node->subreq.oc);
	}
	else {
		TAKE_OBJ_NOTNULL(subreq->objcore, &node->subreq.oc,
		    OBJCORE_MAGIC);
	}

	if (subreq->wrk != NULL)
		assert(subreq->wrk == req->wrk);
	else
		subreq->wrk = req->wrk;

	if (subreq->topreq != NULL)
		assert(subreq->topreq == req);
	else
		subreq->topreq = req;
}

/* ------------------------------------------------------------
 * node cleanup (not holding tree lock)
 *
 * must be idempotent
 */
static void
fini_final(struct req *req, struct node *node)
{
	(void) req;

	assert(node->type == T_FINAL);

	if (node->final.fi_state == FI_DESTROYED)
		return;

	AZ(pthread_mutex_lock(&node->final.fi_mtx));

	if (node->final.fi_state < FI_GO) {
		node->final.fi_state = FI_GO;
		AZ(pthread_cond_signal(&node->final.fi_cond));
	}

	if (node->final.fi_state < FI_DONE)
		AZ(pthread_cond_wait(&node->final.fi_cond,
				     &node->final.fi_mtx));

	AZ(pthread_mutex_unlock(&node->final.fi_mtx));

	assert(node->final.fi_state == FI_DONE);

	AZ(pthread_cond_destroy(&node->final.fi_cond));
	AZ(pthread_mutex_destroy(&node->final.fi_mtx));

	node->final.fi_state = FI_DESTROYED;
}

static void
fini_subreq(struct req *req, struct node *node)
{
	struct req *subreq;
	struct boc *boc = NULL;
	struct pesi *pesi;

	assert(node->type == T_SUBREQ);

	AN(node->subreq.done);

	if (node->subreq.req == NULL) {
		AZ(node->subreq.boc);
		AZ(node->subreq.oc);

		return;
	}

	subreq_fixup(node, req);

	TAKE_OBJ_NOTNULL(subreq, &node->subreq.req, REQ_MAGIC);
	if (node->subreq.boc)
		TAKE_OBJ_NOTNULL(boc, &node->subreq.boc, BOC_MAGIC);
	AZ(node->subreq.oc);

	VDP_close(subreq);	// need before pesi_destroy

	/* bottom of cnt_transmit() */
	HSH_Cancel(req->wrk, subreq->objcore, boc);

	if (boc != NULL)
		HSH_DerefBoc(req->wrk, subreq->objcore);

	(void)HSH_DerefObjCore(req->wrk, &subreq->objcore, 0);

	TAKE_OBJ_NOTNULL(pesi, &subreq->transport_priv, PESI_MAGIC);
	pesi_destroy(&pesi);
	req_fini(&subreq, req->wrk);

	AZ(pthread_cond_destroy(&node->subreq.cond));
}

static void
fini_data(struct req *req, struct node *node)
{
	(void) req;

	assert(node->type == T_DATA);

	if (node->data.st == NULL)
		return;

	stv_transient->sml_free(node->data.st);
	node->data.st = NULL;
}

/* ------------------------------------------------------------
 * node delivery (not holding tree lock)
 */

static int
push_final(struct req *req, struct bytes_tree *tree,
    struct node *node, const struct node *next)
{
	(void) req;
	(void) tree;
	(void) next;

	assert(node->type == T_FINAL);
	assert(node->final.fi_state == FI_READY);

	AZ(pthread_mutex_lock(&node->final.fi_mtx));

	node->final.fi_state = FI_GO;
	AZ(pthread_cond_signal(&node->final.fi_cond));

	if (node->final.fi_state < FI_DONE)
		AZ(pthread_cond_wait(&node->final.fi_cond,
				     &node->final.fi_mtx));

	AZ(pthread_mutex_unlock(&node->final.fi_mtx));

	assert(node->final.fi_state == FI_DONE);
	return (tree->retval);
}

static int
push_subreq(struct req *req, struct bytes_tree *tree,
    struct node *node, const struct node *next)
{
	assert(node->type == T_SUBREQ);

	(void) next;

	if (node->subreq.done == 0) {
		Lck_Lock(&tree->tree_lock);
		if (node->subreq.done == 0)
			AZ(Lck_CondWait(
				   &node->subreq.cond,
				   &tree->tree_lock, 0));
		Lck_Unlock(&tree->tree_lock);
	}

	subreq_fixup(node, req);

	AZ(node->subreq.req->objcore->flags & OC_F_FINAL);

	VSLdbg(req, "DeliverObj from top");
	VDP_DeliverObj(node->subreq.req);
	VDP_close(node->subreq.req);

	// fini_subreq() happening later

	// tree->retval is sort-of inside out here, but
	// that detail really should not matter
	return (tree->retval);
}

static void
gzip_tailbuf(uint8_t tailbuf[GZIP_TAILBUF_SZ], struct nexus_gzip *gz)
{
	AN(gz->is);
	AZ(gz->up);

	tailbuf[0] = 0x01;
	tailbuf[1] = 0x00;
	tailbuf[2] = 0x00;
	tailbuf[3] = 0xff;
	tailbuf[4] = 0xff;

	/* Emit CRC32 */
	vle32enc(tailbuf + 5, gz->crc);

	/* MOD(2^32) length */
	vle32enc(tailbuf + 9, gz->l_crc);
}

static int
push_crc(struct req *req, struct bytes_tree *tree,
    struct node *node, const struct node *next)
{
	struct node *nex;
	struct nexus_gzip *parent_gzip;
	uint8_t tailbuf[GZIP_TAILBUF_SZ];

	assert(node->type == T_CRC);

	(void) req;
	(void) tree;

	nex = node->parent;
	CHECK_OBJ_NOTNULL(nex, NODE_MAGIC);
	assert(nex->type == T_NEXUS);
	AN(nex->nexus.gzip.is);

	switch(node->crc.ctype) {
	case UPDATE:
		nex->nexus.gzip.crc = crc32_combine(
			nex->nexus.gzip.crc, node->crc.icrc,
			node->crc.l_icrc);
		nex->nexus.gzip.l_crc += node->crc.l_icrc;
		break;
	case FINAL:
		parent_gzip = nex->nexus.gzip.up;

		if (next)
			assert(next->parent != nex);
		if (parent_gzip) {
			parent_gzip->crc = crc32_combine(
				parent_gzip->crc, nex->nexus.gzip.crc,
				nex->nexus.gzip.l_crc);
			parent_gzip->l_crc += nex->nexus.gzip.l_crc;
		} else {
			gzip_tailbuf(tailbuf, &nex->nexus.gzip);
			return (VDP_bytes(nex->nexus.req,
			    VDP_FLUSH, tailbuf, sizeof(tailbuf)));
		}
		break;
	default:
		INCOMPL();
	}

	return (0);
}


static int
push_data(struct req *req, struct bytes_tree *tree,
    struct node *node, const struct node *next)
{
	const void *p;
	enum vdp_action act;

	assert(node->type == T_DATA);

	(void) tree;
	(void) req;

	/*
	 * we ignore any act from the node and push all we got
	 * with VDP_FLUSH at the last node
	 */
	act = (next == NULL) ? VDP_FLUSH : VDP_NULL;

	p = node->data.ptr;
	if (p == NULL && node->data.len > 0) {
		CHECK_OBJ_NOTNULL(node->data.st, STORAGE_MAGIC);
		p = node->data.st->ptr;
	}

	AN(node->parent);
	assert(node->parent->type == T_NEXUS);
	return (VDP_bytes(node->parent->nexus.req, act, p, node->data.len));
}

/*
 * deliver unpended notes while not holding the tree lock
 *
 * returns VDP_bytes upon any error
 */
static int
bytes_push_worklist(struct req *req, struct bytes_tree *tree,
    struct node_head *work)
{
	struct node *node, *next, *parent;
	int retval = 0;

	node = VSTAILQ_FIRST(work);
	AN(node);
	while (node) {
		next = VSTAILQ_NEXT(node, unpend);

		// not using assert_node: it needs locking
		assert(node->state == ST_UNPENDING);

		switch (node->type) {
		case T_FINAL:
			retval = push_final(req, tree, node, next);
			break;
		case T_SUBREQ:
			retval = push_subreq(req, tree, node, next);
			break;
		case T_DATA:
			retval = push_data(req, tree, node, next);
			break;
		case T_CRC:
			retval = push_crc(req, tree, node, next);
			break;
		default:
			INCOMPL();
		}

		if (retval)
			break;

		if (VSTAILQ_NEXT(node, sibling) == NULL) {
			parent = node->parent;
			if (parent->parent == NULL)
				(void) 0;
			else if (parent->state < ST_CLOSED)
				assert(next == NULL);
			else
				VDP_close(parent->nexus.req);
		}
		node = next;
	}

	if (retval) {
		(void) VDP_bytes(req, VDP_FLUSH, NULL, 0);
		return (retval);
	}

	/* not running the finis in the error case, they
	 * will be called from final cleanup via tree_free()
	 *
	 * this avoids waiting for subreq tasks
	 */

	VSTAILQ_FOREACH(node, work, unpend) {
		node_fini(req, node);

		if (VSTAILQ_NEXT(node, sibling) != NULL)
			continue;

		parent = node->parent;
		/* don't touch the root */
		if (parent->parent == NULL)
			continue;

		if (parent->state < ST_CLOSED)
			continue;

		/*
		 * fini the now unneeded parent ASAP
		 *
		 * this is separate from set_delived() because
		 * - set_delivered() happens under the lock
		 * - set_delivered() can also be called from
		 *   within the parent's FSM
		 */
		node_fini(req, parent);
	}

	return (retval);
}

static void
bytes_set_delivered_worklist(struct bytes_tree *tree,
    struct node_head *work)
{
	struct node *node;

	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);
	Lck_AssertHeld(&tree->tree_lock);

	VSTAILQ_FOREACH(node, work, unpend)
		set_delivered(tree, node);
}

static void
bytes_unpend(struct req *req, struct bytes_tree *tree)
{
	struct node_head work;
	int retval;

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(req->wrk, WORKER_MAGIC);

	if (tree->unpend_owner != NULL)
		return;

	tree->unpend_owner = req->wrk;

	do {
		Lck_AssertHeld(&tree->tree_lock);

		assert(tree->unpend_owner == req->wrk);

		VSTAILQ_INIT(&work);

		bytes_unpend_worklist(req, tree, &work);
		if (VSTAILQ_EMPTY(&work) || tree->retval)
			break;

		Lck_Unlock(&tree->tree_lock);
		retval = bytes_push_worklist(req, tree, &work);
		Lck_Lock(&tree->tree_lock);

		if (retval) {
			tree->retval = retval;
			break;
		}
		bytes_set_delivered_worklist(tree, &work);
	} while (1);

	tree->unpend_owner = NULL;
}

static int v_matchproto_(vdp_init_f)
pesi_buf_init(struct req *req, void **priv)
{
	(void)req;
	(void)priv;
	return (0);
}

static int v_matchproto_(vdp_fini_f)
pesi_buf_fini(struct req *req, void **priv)
{
	(void)req;

	*priv = NULL;
	return (0);
}

static int v_matchproto_(vdp_bytes_f)
pesi_buf_bytes(struct req *req, enum vdp_action act, void **priv,
	       const void *ptr, ssize_t len)
{
	struct bytes_tree *tree;
	struct node *node, *parent;
	struct pesi *pesi;
	unsigned refok;

	CAST_OBJ_NOTNULL(pesi, *priv, PESI_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pesi_tree, PESI_TREE_MAGIC);
	tree = pesi->pesi_tree->tree;
	CHECK_OBJ_NOTNULL(tree, BYTES_TREE_MAGIC);

	// magic pointers in use, so no check for len == 0
	if (tree->retval || ptr == NULL)
		return (tree->retval);

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	parent = pesi->node;
	CHECK_OBJ_NOTNULL(parent, NODE_MAGIC);
	assert(parent->type == T_NEXUS);
	VSLdbgv(req, "bytes_add: parent=%p front=%p",
		parent, tree->front);

	assert(req == parent->nexus.req);

	VSLdbg(req, "bytes_add: adding data to node");
	node = node_alloc();
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	node->type = T_DATA;
	node->state = ST_DATA;
	node->data.len = len;
	node->data.act = act;

	if (len == 0) {
		// may be a magic pointer
		AZ(node->data.st);
		node->data.ptr = ptr;

		node_insert(tree, parent, node);

		VSLdbg(req, "bytes_add: exit len == 0");
		return (tree->retval);
	}

	/*
	 * non-api knowledge from VDP_DeliverObj()
	 *  final = req->objcore->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP);
	 */

	/*
	 * if data is coming from a varnish-cache simple storage, we know that
	 * - for non-final objects
	 * - as long as we keep a reference
	 *
	 * the storage will not go away, so any vector we get pushed will remain
	 * valid. We only do this check once
	 *
	 */
	if ((node->type == T_NEXUS) &&
	    (req->objcore->flags & OC_F_FINAL) == 0 &&
	    req->objcore->stobj->stevedore->methods == &SML_methods) {
		AZ(parent->nexus.oc);

		parent->nexus.oc = req->objcore;
		HSH_Ref(parent->nexus.oc);
	}

	// XXX make gzip_hdr a T_CRC ?
	refok = ((parent->nexus.oc != NULL) ||
		 // from vdp_pesi_bytes
		 (ptr == gzip_hdr));


	AN(ptr);
	AZ(node->data.ptr);
	if (refok) {
		VSLdbg(req, "bytes_add: stv ref");
		node->data.ptr = ptr;
	}
	else {
		VSLdbg(req, "bytes_add: allocating from transient");
		node->data.st = stv_transient->sml_alloc(stv_transient, len);
		if (node->data.st == NULL) {
			VSLb(req->vsl, SLT_Error, "Cannot allocate transient "
			     "storage to buffer response data while waiting "
			     "for parallel ESI");
			VSLdbg(req, "bytes_add: exit after transient alloc "
			       "failure");

			FREE_OBJ(node);

			Lck_Lock(&tree->tree_lock);
			if (tree->retval == 0)
				tree->retval = -1;
			Lck_Unlock(&tree->tree_lock);

			return (tree->retval);
		}
		else {
			node->data.act = VDP_NULL;
			AZ(node->data.st->len);
			assert(node->data.st->space == len);
			memcpy(node->data.st->ptr, ptr, len);
		}
	}

	node_insert(tree, parent, node);

	/*
	 * check unlocked if an attempt to unpend makes any sense at all
	 */
	AN(tree->front);
	if (pesi->flags & PF_CFG_FRONT_PUSH &&
	    tree->unpend_owner != NULL &&
	    (tree->front->parent == NULL ||
	     tree->front == parent ||
	     tree->front->parent == parent ||
	     tree->front_owner == req->wrk)) {
		VSLdbg(req, "front push");

		// XXX for this node, we can spare buffering if got pushed
		// XXX double locking with node_insert

		Lck_Lock(&tree->tree_lock);
		if (parent->state == ST_PRIVATE)
			set_open(tree, parent, req->wrk);
		assert(parent->state = ST_OPEN);
		bytes_unpend(req, tree);
		Lck_Unlock(&tree->tree_lock);
	}

	VSLdbgv(req, "bytes_add to %s parent: exit",
	    parent->state == ST_PRIVATE ? "private" : "open");

	return (tree->retval);
}

const struct vdp VDP_pesi_buf = {
	.name =		"pesi_buf",
	.init =		pesi_buf_init,
	.bytes =	pesi_buf_bytes,
	.fini =		pesi_buf_fini,
};

/* VDP pesi */

static int
push_vdps(struct req *req, struct nexus_gzip *gz)
{
	int i;
	struct vped_gzgz_priv *vgzgz;

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);
	CHECK_OBJ_NOTNULL(req->wrk, WORKER_MAGIC);

#ifdef DEBUG
	/* XXX 5-28 */
	struct vdp_entry *vdpe;
	CHECK_OBJ_NOTNULL(req->vdc, VDP_CTX_MAGIC);
	VTAILQ_FOREACH(vdpe, &req->vdc->vdp, list) {
		CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
		VSLdbgv(req,
			"push_vdps: VDP list before pushing parent VDPs: %s",
			vdpe->vdp->name);
	}
#endif

	i = ObjCheckFlag(req->wrk, req->objcore, OF_GZIPED);
	if (gz->is && i && !(req->res_mode & RES_ESI)) {
		/* A gzip'ed include which is not ESI processed */

		VSLdbg(req, "push_vdps: pushing gzgz");
		if ((vgzgz = WS_Alloc(req->ws, sizeof(*vgzgz))) == NULL) {
			VSLb(req->vsl, SLT_Error,
			     "Insufficient workspace for ESI gzip data");
			return (-1);
		}
		INIT_OBJ(vgzgz, VPED_GZGZ_PRIV_MAGIC);
		vgzgz->gz = gz;
		AZ(VDP_Push(req, &vped_gzgz, vgzgz));
	}
	else if (gz->is && !i) {
		VSLdbg(req, "push_vdps: pushing pretend_gz");
		/* Non-Gzip'ed include in gzip'ed parent */
		AZ(VDP_Push(req, &vped_pretend_gz, gz));
	}
	// else just push to parent

#ifdef DEBUG
	/* XXX debugging */
	CHECK_OBJ_NOTNULL(req->vdc, VDP_CTX_MAGIC);
	VTAILQ_FOREACH(vdpe, &req->vdc->vdp, list) {
		CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
		VSLdbgv(req, "push_vdps: VDP list: %s", vdpe->vdp->name);
	}
#endif
	return (0);
}

static int v_matchproto_(vdp_init_f)
vdp_pesi_init(struct req *req, void **priv)
{
	struct pesi *pesi;
	struct pecx *pecx;
	struct pesi_tree *pesi_tree;

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	AN(priv);

	if (!ObjHasAttr(req->wrk, req->objcore, OA_ESIDATA))
		return (1);

	if (req->esi_level > 0) {
		if (*priv != NULL) {
			CHECK_OBJ((struct pesi *)*priv, PESI_MAGIC);
			assert(*priv == req->transport_priv);
			WS_Assert_Allocated(req->ws, *priv,
					    sizeof(struct pesi));
			return (0);
		}

		if (VALID_OBJ((struct pesi *)req->transport_priv, PESI_MAGIC)) {
			WS_Assert_Allocated(req->ws, req->transport_priv,
					    sizeof(struct pesi));
			*priv = req->transport_priv;
			/* XXX debugging */
			struct vdp_entry *vdpe;
			CHECK_OBJ_NOTNULL(req->vdc, VDP_CTX_MAGIC);
			VTAILQ_FOREACH(vdpe, &req->vdc->vdp, list) {
				CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
				VSLdbgv(req, "vdp_pesi_init: VDP list: %s",
					vdpe->vdp->name);
			}
			return (0);
		}
	}

	AZ(req->esi_level);

	if ((pesi_tree = WS_Alloc(req->ws, sizeof(*pesi_tree))) == NULL) {
		VSLb(req->vsl, SLT_Error, "insufficient workspace for "
		     "global parallel ESI data");
		return (-1);
	}

	INIT_OBJ(pesi_tree, PESI_TREE_MAGIC);

	if ((pesi_tree->tree = WS_Alloc(req->ws, sizeof(*pesi_tree->tree)))
	    == NULL) {
		VSLb(req->vsl, SLT_Error, "insufficient workspace for "
		     "parallel ESI tree data");
		return (-1);
	}
	INIT_OBJ(pesi_tree->tree, BYTES_TREE_MAGIC);

	Lck_New(&pesi_tree->task_lock, lck_pesi_tree);
	AZ(pthread_cond_init(&pesi_tree->task_cond, NULL));
	VTAILQ_INIT(&pesi_tree->task_head);
	AZ(pesi_tree->task_running);

	pesi = pesi_new(req->ws, pesi_tree);
	if (pesi == NULL) {
		VSLb(req->vsl, SLT_Error,
		     "Cannot allocate workspace for parallel ESI data");
		return (-1);
	}
	pecx = pesi->pecx;
	*priv = pesi;
	RFC2616_Weaken_Etag(req->resp);
	req->res_mode |= RES_ESI;
	if (req->resp_len != 0)
		req->resp_len = -1;

	pesi_tree->tree->root = node_alloc();
	CHECK_OBJ_NOTNULL(pesi_tree->tree->root, NODE_MAGIC);
	pesi_tree->tree->front = pesi_tree->tree->root;
	pesi_tree->tree->root->state = ST_PRIVATE;
	pesi_tree->tree->root->type = T_NEXUS;
	pesi_tree->tree->root->nexus.req = req;
	pesi_tree->tree->root->nexus.gzip.magic = NEXUS_GZIP_MAGIC;
	VSTAILQ_INIT(&pesi_tree->tree->root->nexus.children);
	Lck_New(&pesi_tree->tree->tree_lock, lck_bytes_tree);
	pesi_tree->tree->npending = 1;
	AZ(pthread_cond_init(&pesi_tree->tree->cond, NULL));
	AZ(pesi_tree->tree->retval);

	pesi->node = pesi_tree->tree->root;
	AZ(pecx->state);
	AZ(pesi->woken);

	VDP_Push(req, &VDP_pesi_buf, pesi);
	return (0);
}

static int v_matchproto_(vdp_fini_f)
vdp_pesi_fini(struct req *req, void **priv)
{
	struct pesi *pesi;
	struct pesi_tree *pesi_tree;
	struct bytes_tree *bytes_tree;
	struct node *node;

	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	VSLdbg(req, "vdp_pesi_fini: enter");
	AN(priv);

	CAST_OBJ_NOTNULL(pesi, *priv, PESI_MAGIC);
	pesi_tree = pesi->pesi_tree;
	CHECK_OBJ_NOTNULL(pesi_tree, PESI_TREE_MAGIC);
	bytes_tree = pesi_tree->tree;
	CHECK_OBJ_NOTNULL(bytes_tree, BYTES_TREE_MAGIC);

	VSLdbgv(req, "vdp_pesi_fini: esi_level=%u", req->esi_level);

	node = pesi->node;
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	if (req->esi_level > 0) {
		assert(req->transport_priv == pesi);
		*priv = NULL;
		req->transport_priv = NULL;

		pesi_destroy(&pesi);

		return (0);
	}

	assert(req->transport_priv == NULL ||
	       *(unsigned *)req->transport_priv != PESI_MAGIC);

	/*
	 * at this point, the tree must be delivered completely, but
	 * tasks our shared data structures could still be running
	 */
	if (bytes_tree->retval) {
		assert_node(bytes_tree->root, CHK_ANY);
	}
	else {
		assert(bytes_tree->npending == 0);
		assert_node(bytes_tree->root, CHK_DELI);
	}

	CHECK_OBJ_NOTNULL(pesi_tree, PESI_TREE_MAGIC);
	assert(pesi_tree == pesi->pesi_tree);
	CHECK_OBJ_NOTNULL(bytes_tree, BYTES_TREE_MAGIC);
	assert(bytes_tree == pesi_tree->tree);

	pesi_destroy(&pesi);
	task_fini(pesi_tree, pesi);

	/*
	 * ensure that all tasks are done
	 *
	 * there can be at most one signal ever being posted on task_cond, so,
	 * deliberately, this is if (...) not while (...)
	 */
	if (pesi_tree->task_running > 0) {
		Lck_Lock(&pesi_tree->task_lock);
		if (pesi_tree->task_running > 0)
			AZ(Lck_CondWait(&pesi_tree->task_cond,
					&pesi_tree->task_lock, 0));
		Lck_Unlock(&pesi_tree->task_lock);
	}

	/*
	 * for an early client close, our subrequests had no chance to run
	 *
	 * we likely need more error handling
	 */
	tree_fini(req, bytes_tree->root);

	AZ(pesi_tree->task_running);
	assert(VTAILQ_EMPTY(&pesi_tree->task_head));
	Lck_Delete(&pesi_tree->task_lock);
	AZ(pthread_cond_destroy(&pesi_tree->task_cond));

	Lck_Delete(&bytes_tree->tree_lock);
	AZ(pthread_cond_destroy(&bytes_tree->cond));

	tree_free(req, bytes_tree->root);

	*priv = NULL;
	return (0);
}

static int v_matchproto_(vdp_bytes_f)
vdp_pesi_bytes(struct req *req, enum vdp_action act, void **priv,
	       const void *ptr, ssize_t len)
{
	uint8_t *q, *r;
	ssize_t l = 0;
	const uint8_t *pp;
	struct pesi *pesi;
	struct pecx *pecx;
	struct nexus_gzip *parent_gzip = NULL;
	struct bytes_tree *tree;
	struct node *node, *child = NULL;
	int retval = 0, parallel;

	AN(priv);
	CAST_OBJ_NOTNULL(pesi, *priv, PESI_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pecx, PECX_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pesi_tree, PESI_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pesi_tree->tree, BYTES_TREE_MAGIC);
	pecx = pesi->pecx;
	tree = pesi->pesi_tree->tree;

	node = pesi->node;
	CHECK_OBJ_NOTNULL(node, NODE_MAGIC);

	/*
	 * special case for ESI with http_gzip_support off: gzip tail getting
	 * pushed to us after the state engine has finished
	 *
	 * we would want to remove ourselves from the VDP list, but cannot, as
	 * that would trigger our _fini while we are in _bytes - catch22...
	 *
	 * see e00031.vtc and e00033.vtc
	 */
	if (req->esi_level == 0 &&
	    node->state == ST_DELIVERED &&
	    pecx->state == 99) {
		return (VDP_bytes(req, act, ptr, len));
	}

	assert(node->type == T_NEXUS);
	assert(node->state == ST_OPEN || node->state == ST_PRIVATE);
	assert(ObjHasAttr(req->wrk, req->objcore, OA_ESIDATA));

	if (node->parent != NULL && node->parent->nexus.gzip.is)
		parent_gzip = &node->parent->nexus.gzip;

	pp = ptr;
	while (1) {
		switch (pecx->state) {
		case 0:
			pecx->p = ObjGetAttr(req->wrk, req->objcore,
			    OA_ESIDATA, &l);
			AN(pecx->p);
			assert(l > 0);
			pecx->e = pecx->p + l;

			if (*pecx->p == VEC_GZ) {
				/*
				 * this will, in pesi_buf_bytes(), create a
				 * T_DATA node which ref's gzip_hdr. Should we
				 * rather add a T_GZIPHDR node?
				 */
				if (parent_gzip == NULL)
					retval = VDP_bytes(req, VDP_NULL,
							   gzip_hdr, 10);
				node->nexus.gzip.is = 1;
				node->nexus.gzip.crc = crc32(0L, Z_NULL, 0);
				AZ(node->nexus.gzip.l_crc);
				node->nexus.gzip.up = parent_gzip;

				pecx->p++;
			}
			pecx->state = 1;
			break;
		case 1:
			if (pecx->p >= pecx->e) {
				pecx->state = 2;
				break;
			}
			switch (*pecx->p) {
			case VEC_V1:
			case VEC_V2:
			case VEC_V8:
				pecx->l = ved_decode_len(req, &pecx->p);
				if (pecx->l < 0)
					return (-1);
				Debug("VEC_V(%d)\n", (int)pecx->l);
				if (node->nexus.gzip.is) {
					assert(*pecx->p == VEC_C1 ||
					    *pecx->p == VEC_C2 ||
					    *pecx->p == VEC_C8);
					l = ved_decode_len(req, &pecx->p);
					if (l < 0)
						return (-1);

					AZ(child);
					child = node_alloc();
					CHECK_OBJ_NOTNULL(child, NODE_MAGIC);
					child->type = T_CRC;
					child->state = ST_DATA;
					child->crc.ctype = UPDATE;
					child->crc.icrc = vbe32dec(pecx->p);
					child->crc.l_icrc = l;
					node_insert(tree, node, child);
					child = NULL;

					pecx->p += 4;
				}
				pecx->state = 3;
				break;
			case VEC_S1:
			case VEC_S2:
			case VEC_S8:
				pecx->l = ved_decode_len(req, &pecx->p);
				if (pecx->l < 0)
					return (-1);
				Debug("SKIP1(%d)\n", (int)pecx->l);
				pecx->state = 4;
				break;
			case VEC_INCL:
				pecx->p++;
				q = (void*)strchr((const char*)pecx->p, '\0');
				AN(q);
				q++;
				r = (void*)strchr((const char*)q, '\0');
				AN(r);
				Debug("INCL [%s][%s] BEGIN\n", q, pecx->p);

				AZ(child);
				child = node_alloc();
				CHECK_OBJ_NOTNULL(child, NODE_MAGIC);
				child->type = T_NEXUS;
				child->state = ST_PRIVATE;
				child->nexus.gzip.magic = NEXUS_GZIP_MAGIC;

				node_insert(tree, node, child);

				VSLdbg(req, "ved_vdp: call ved_include");
				parallel =
					vped_include(req, (const char*)q,
						     (const char*)pecx->p,
						     pesi, child);
				child = NULL;
				VSLdbgv(req, "ved_vdp: vped_include()=%d",
					parallel);
				assert(parallel >= 0);
				Debug("INCL [%s][%s] END\n", q, pecx->p);
				pecx->p = r + 1;
				break;
			default:
				VSLb(req->vsl, SLT_Error,
				    "ESI corruption line %d 0x%02x [%s]\n",
				    __LINE__, *pecx->p, pecx->p);
				WRONG("ESI-codes: Illegal code");
			}
			break;
		case 2:
			if (node->nexus.gzip.is) {
				AZ(child);
				child = node_alloc();
				CHECK_OBJ_NOTNULL(child, NODE_MAGIC);
				child->type = T_CRC;
				child->state = ST_DATA;
				child->crc.ctype = FINAL;
				node_insert(tree, node, child);
				child = NULL;
			}
			assert(pecx->p >= pecx->e);

			pecx->state = 99;
			break;
		case 3:
		case 4:
			/*
			 * There is no guarantee that the 'l' bytes are all
			 * in the same storage segment, so loop over storage
			 * until we have processed them all.
			 */
			VSLdbgv(req, "ved_vdp: pecx->state=%d", pecx->state);
			if (pecx->l <= len) {
				if (pecx->state == 3)
					retval = VDP_bytes(req, act, pp, pecx->l);
				len -= pecx->l;
				pp += pecx->l;
				pecx->state = 1;
				break;
			}
			if (pecx->state == 3 && len > 0)
				retval = VDP_bytes(req, act, pp, len);
			pecx->l -= len;
			return (tree->retval);
		case 99:
			/*
			 * VEP does not account for the PAD+CRC+LEN
			 * so we can see up to approx 15 bytes here.
			 */

			if (req->esi_level > 0)
				return (tree->retval);

			/*
			 * we need to wait for completion here because at vdp
			 * fini time, the V1D is already closed
			 */
			Lck_Lock(&tree->tree_lock);
			set_closed(tree, node, req->wrk);

			assert_vdp_next_not(req, &VDP_pesi);
			vped_close_vdp(req, 1, &VDP_pesi_buf);
			assert_vdp_next_not(req, &VDP_pesi);

			bytes_unpend(req, tree);
			while (!tree->retval
			       && (tree->root->state != ST_DELIVERED ||
				   tree->npending > 0)) {
				VSLdbgv(req, "vdp_pesi_bytes: waiting - "
					"state=%d npending=%d",
					tree->root->state,
					tree->npending);
#if 1
				AZ(Lck_CondWait(&tree->cond,
						&tree->tree_lock, 0));
#else
				(void)(Lck_CondWait(&tree->cond,
						    &tree->tree_lock,
						    VTIM_real() + 2));
				AZ(errno);
#endif
				bytes_unpend(req, tree);
			}
			Lck_Unlock(&tree->tree_lock);

			return (tree->retval);
		default:
			WRONG("FOO");
			break;
		}
		if (retval)
			return (retval);
		if (tree->retval)
			return (tree->retval);
	}
}

const struct vdp VDP_pesi = {
	.name =		"pesi",
	.init =		vdp_pesi_init,
	.bytes =	vdp_pesi_bytes,
	.fini =		vdp_pesi_fini,
};

/* Transport interface */

static int v_matchproto_(vtr_minimal_response_f)
vped_minimal_response(struct req *req, uint16_t status)
{
	(void)req;
	(void)status;
	WRONG("esi:includes should not try minimal responses");
}

static void v_matchproto_(vtr_reembark_f)
vped_reembark(struct worker *wrk, struct req *req)
{
	struct pesi *pesi;

	(void)wrk;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CAST_OBJ_NOTNULL(pesi, req->transport_priv, PESI_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pecx, PECX_MAGIC);

	VSLb(req->vsl, SLT_Debug, "reembark wakeup");
	Lck_Lock(&req->sp->mtx);
	pesi->woken = 1;
	AZ(pthread_cond_signal(&pesi->wrk->cond));
	Lck_Unlock(&req->sp->mtx);
}

static void
assert_vdp_next_not(struct req *req, const struct vdp *vdp)
{
	struct vdp_entry *nxt;
	struct vdp_ctx *vdc;

	vdc = req->vdc;
	nxt = vdc->nxt;

	CHECK_OBJ_NOTNULL(nxt, VDP_ENTRY_MAGIC);
	assert(nxt->vdp != vdp);
}

/*
 * close first vdp and assert it has the right type
 *
 * there must be at least one vdp remaining
 */
static void
vped_close_vdp(struct req *req, int skip, const struct vdp *vdp)
{
	struct vdp_entry *vdpe, *nxt;
	struct vdp_ctx *vdc;

	vdc = req->vdc;
	nxt = vdc->nxt;
	vdpe = VTAILQ_FIRST(&vdc->vdp);

	/*
	 * it is a caller error to skip too many vdps, so if any of these checks
	 * fail, know your stack...
	 */
	while (skip-- > 0) {
		CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);
		vdpe = VTAILQ_NEXT(vdpe, list);
	}
	CHECK_OBJ_NOTNULL(vdpe, VDP_ENTRY_MAGIC);

	assert(vdpe->vdp == vdp);
	AN(vdpe->vdp->fini);
	AZ(vdpe->vdp->fini(req, &vdpe->priv));
	AZ(vdpe->priv);
	if (nxt == vdpe)
		nxt = VTAILQ_NEXT(vdpe, list);
	VTAILQ_REMOVE(&vdc->vdp, vdpe, list);

	if (nxt == NULL)
		nxt = VTAILQ_FIRST(&vdc->vdp);
	CHECK_OBJ_NOTNULL(nxt, VDP_ENTRY_MAGIC);
	vdc->nxt = nxt;
}

static void v_matchproto_(vtr_deliver_f)
vped_deliver(struct req *req, struct boc *boc, int wantbody)
{
	int i, is_esi, is_final;
	struct pesi *pesi;
	struct bytes_tree *tree;
	struct node *node, *parent;

	VSLdbgv(req, "vped_deliver: req=%p boc=%p wantbody=%d", req, boc,
		wantbody);
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
	CHECK_OBJ_NOTNULL(req->objcore, OBJCORE_MAGIC);

	CAST_OBJ_NOTNULL(pesi, req->transport_priv, PESI_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pesi_tree, PESI_TREE_MAGIC);
	CHECK_OBJ_NOTNULL(pesi->pesi_tree->tree, BYTES_TREE_MAGIC);
	tree = pesi->pesi_tree->tree;

	node = pesi->node;
	parent = node->parent;
	CHECK_OBJ_NOTNULL(parent, NODE_MAGIC);
	assert(parent->type == T_NEXUS);

	if (wantbody == 0) {
		VDP_close(req);
		return;
	}

	VSLdbgv(req, "vped_deliver: ObjGetLen=%lu",
		ObjGetLen(req->wrk, req->objcore));
	if (boc == NULL && ObjGetLen(req->wrk, req->objcore) == 0) {
		VDP_close(req);
		return;
	}

	is_esi = ObjHasAttr(req->wrk, req->objcore, OA_ESIDATA);
	is_final = (req->objcore->flags & OC_F_FINAL) != 0;

	i = ObjCheckFlag(req->wrk, req->objcore, OF_GZIPED);

	if (parent->nexus.gzip.is && i && !(req->res_mode & RES_ESI)) {
		/* OA_GZIPBITS are not valid until BOS_FINISHED */
		if (boc != NULL)
			ObjWaitState(req->objcore, BOS_FINISHED);

		if (req->objcore->flags & OC_F_FAILED) {
			/* No way of signalling errors in the middle of
			   the ESI body. Omit this ESI fragment. */
			VDP_close(req);
			return;
		}
	}

	if (is_final && !is_esi && (pesi->flags & PF_CFG_BLOCK_FINAL) == 0) {
		VSLdbg(req, "vped_deliver: T_NEXUS buffering no ESI");

		VDP_Push(req, &VDP_pesi_buf, pesi);
		AN(parent);
		assert(parent->type == T_NEXUS);
		AZ(push_vdps(req, &parent->nexus.gzip));
		AZ(VDP_Push(req, &vped_to_parent, parent->nexus.req));

		(void)VDP_DeliverObj(req);

		/* pesi_buf does not pesi_destroy() on fini */
		req->transport_priv = NULL;
		pesi_destroy(&pesi);
		vped_close_vdp(req, 0, &VDP_pesi_buf);
		return;
	}
	else if (is_final && !is_esi) {
		VSLdbg(req, "vped_deliver: T_FINAL");

		AZ(push_vdps(req, &parent->nexus.gzip));
		AZ(VDP_Push(req, &vped_to_parent, parent->nexus.req));

		node_mutate_prep(tree, node);

		AZ(pthread_mutex_init(&node->final.fi_mtx, NULL));
		AZ(pthread_cond_init(&node->final.fi_cond, NULL));
		assert(node->final.fi_state == FI_READY);

		node_mutate_lock(tree, node, T_FINAL, ST_DATA);

		/*
		 * we deliberately take the fi_mtx under the tree mutex (via
		 * node_mutate_lock) to ensure unpending never gets unlocked
		 * access to ensure we get the signal
		 */
		AZ(pthread_mutex_lock(&node->final.fi_mtx));

		node_mutate_unlock(tree);

		if (node->final.fi_state < FI_GO)
			AZ(pthread_cond_wait(&node->final.fi_cond,
					     &node->final.fi_mtx));

		assert(node->final.fi_state == FI_GO);

		(void)VDP_DeliverObj(req);

		VDP_close(req);

		node->final.fi_state = FI_DONE;
		AZ(pthread_cond_signal(&node->final.fi_cond));
		AZ(pthread_mutex_unlock(&node->final.fi_mtx));

		return;
	}
	else if (! is_esi) {
		VSLdbg(req, "vped_deliver: T_SUBREQ");

		AZ(req->objcore->flags & OC_F_FINAL);
		AZ(push_vdps(req, &parent->nexus.gzip));
		AZ(VDP_Push(req, &vped_to_parent, parent->nexus.req));

		node_mutate_prep(tree, node);

		/* our caller, cnt_transmit(), releases the refs to boc and oc
		 * when we return, so in order to hand this request to the
		 * topreq thread, we need to gain one more
		 */
		node->subreq.req = req;
		node->subreq.boc = HSH_RefBoc(req->objcore);
		HSH_Ref(req->objcore);
		node->subreq.oc = req->objcore;

		AZ(pthread_cond_init(&node->subreq.cond, NULL));

		node_mutate_lock(tree, node, T_SUBREQ, ST_DATA);
		node_mutate_unlock(tree);

		/*
		 * to be done in the other thread:
		 * - VDP_DeliverObj()
		 * - VDP_close()
		 *
		 * from ved_task():
		 * - req_fini()
		 */
		return;
	}

	VSLdbg(req, "vped_deliver: ESI");

	VDP_Push(req, &VDP_pesi_buf, pesi);
	AZ(push_vdps(req, &node->nexus.gzip));
	AN(parent);
	AZ(VDP_Push(req, &vped_to_parent, parent->nexus.req));

	(void)VDP_DeliverObj(req);

	/*
	 * only close the pesi VDPs, the others are to run from topreq delivery
	 */
	vped_close_vdp(req, 0, &VDP_pesi);
	vped_close_vdp(req, 0, &VDP_pesi_buf);
	return;
}

static const struct transport VPED_transport = {
	.magic =		TRANSPORT_MAGIC,
	.name =			"PESI_INCLUDE",
	.deliver =		vped_deliver,
	.reembark =		vped_reembark,
	.minimal_response =	vped_minimal_response,
};

/* VMOD functions */

static volatile struct poolparam poolparam = {
	.min_pool = 10,
	.max_pool = 100,
	.max_age = 10,
};

VCL_VOID
vmod_pool(VRT_CTX, VCL_INT min, VCL_INT max, VCL_DURATION max_age)
{
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	/* cf. tweak_poolparam() in mgt_param_tweak.c */
	if (min <= 0) {
		VFAIL(ctx, "min (%jd) must be > 0 in pool()", min);
		return;
	}
	if (max <= 0) {
		VFAIL(ctx, "max (%jd) must be > 0 in pool()", max);
		return;
	}
	if (max < min) {
		VFAIL(ctx, "max (%jd) < min (%jd) in pool()", max, min);
		return;
	}
	if (max_age < 0.) {
		VFAIL(ctx, "max_age (%.0fs) < 0s in pool()", max_age);
		return;
	}
	if (max_age > 1e6) {
		VFAIL(ctx, "max_age (%.0fs) out of range in pool() (max 10^6s)",
		      max_age);
		return;
	}

	poolparam.min_pool = min;
	poolparam.max_pool = max;
	poolparam.max_age = max_age;
}

/*
 * pesi_filter_on_ws() and pesi_resp_default_filter_list()
 * taken from cache_vrt_filter.c
 *
 */

typedef void pesi_filter_list_t(void *, struct vsb *vsb);

static const char *
pesi_filter_on_ws(struct ws *ws, pesi_filter_list_t *func, void *arg)
{
	unsigned u;
	struct vsb vsb[1];

	AN(func);
	AN(arg);
	u = WS_ReserveAll(ws);
	if (u == 0) {
		WS_Release(ws, 0);
		WS_MarkOverflow(ws);
		return (NULL);
	}
	AN(VSB_new(vsb, ws->f, u, VSB_FIXEDLEN));
	func(arg, vsb);
	if (VSB_finish(vsb)) {
		WS_Release(ws, 0);
		WS_MarkOverflow(ws);
		return (NULL);
	}
	if (VSB_len(vsb)) {
		WS_Release(ws, VSB_len(vsb) + 1);
		return (VSB_data(vsb) + 1);
	}
	WS_Release(ws, 0);
	return ("");
}

static void v_matchproto_(pesi_filter_list_t)
pesi_resp_default_filter_list(void *arg, struct vsb *vsb)
{
	struct req *req;

	CAST_OBJ_NOTNULL(req, arg, REQ_MAGIC);

	// XXX resp_len check does not work for busy objects
	// pesi will still DTRT if the response really is 0

	if (!req->disable_esi && /* req->resp_len != 0 && */
	    ObjHasAttr(req->wrk, req->objcore, OA_ESIDATA))
		VSB_cat(vsb, " pesi");

	if (cache_param->http_gzip_support &&
	    ObjCheckFlag(req->wrk, req->objcore, OF_GZIPED) &&
	    !RFC2616_Req_Gzip(req->http))
		VSB_cat(vsb, " gunzip");

	if (cache_param->http_range_support &&
	    http_GetStatus(req->resp) == 200 &&
	    http_GetHdr(req->http, H_Range, NULL))
		VSB_cat(vsb, " range");
}

VCL_VOID
vmod_activate(VRT_CTX)
{
	struct req *req;
	const char *filters;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (ctx->method != VCL_MET_DELIVER) {
		VRT_fail(ctx, "pesi.activate() may only be called "
		    "from vcl_deliver{}");
		return;
	}

	req = ctx->req;
	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);

	filters = pesi_filter_on_ws(req->ws,
	    pesi_resp_default_filter_list, req);

	if (filters == NULL)
		WS_MarkOverflow(req->ws);
	else
		req->filter_list = filters;
}

VCL_VOID
vmod_serial(VRT_CTX, VCL_BOOL b)
{
	struct vmod_priv *task;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	if (ctx->method != VCL_MET_DELIVER) {
		VRT_fail(ctx, "pesi.serial() may only be called "
		    "from vcl_deliver{}");
		return;
	}

	task = VRT_priv_task(ctx, priv_task_id_serial);

	if (task == NULL) {
		VRT_fail(ctx, "no priv_task");
		return;
	}

	/* for now, just put the flags in the priv */
	assert(sizeof task->priv >= sizeof(unsigned));
	task->priv = b ? (void *)PF_CFG_SERIAL : (void *)0;
}

/* Event function */

/*
 * MPL embeds its own struct memitem into the allocation and blindly reduces the
 * requested size by the memitem size.
 *
 * Yet it fails to expose the struct memitem size such that any client to that
 * API requiring a fixed size cannot directly determine the required allocation
 * size
 *
 * So we request a memitem and determine the struct memitem size indirectly
 *
 * as long as sizeof(struct node) + sizeof(struct memitem) > MEMITEM_GUESS
 * + sizeof (struct memitem), the too small memitem will be freed again and all
 * memitems will be exactly the right size. Otherwise the too large memitem will
 * eventually die for timeout, but the overhead does not matter anyway
 *
 * (gdb) print sizeof(struct memitem)
 * $1 = 32
 */
#define MEMITEM_GUESS 32

static struct mempool *
mpl_init()
{
	struct mempool *mpl;
	unsigned sz;
	void *test;

	node_alloc_sz = MEMITEM_GUESS * 2;
	mpl = MPL_New("pesi", &poolparam, &node_alloc_sz);

	test = MPL_Get(mpl, &sz);
	AN(test);
	assert(sz <= node_alloc_sz);	// no unsigned underflow

	node_alloc_sz += sizeof(struct node);
	node_alloc_sz -= sz;
	MPL_Free(mpl, test);

	return (mpl);
}

/*
 * ... and another shortcoming of MPL: mpl uses a guard thread, yet
 * MPL_Destroy() does not wait for it to finish. So after MPL_Destroy() has
 * returned the struct mempool will still be accessed by the guard thread.
 *
 * we got no other option than to wait for that thread to hopefully have
 * finished.
 *
 * sigh...
 */

static void
mpl_fini(struct mempool **mplp)
{
	MPL_Destroy(mplp);
	VTIM_sleep(0.814 * 2); /* max mpl_slp in MPL code * 2 */
}

int v_matchproto_(vmod_event_f)
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	ASSERT_CLI();
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);

	switch (e) {
	case VCL_EVENT_LOAD:
		if (loadcnt++ == 0) {
			AZ(vsc_seg);
			lck_bytes_tree = Lck_CreateClass(&vsc_seg,
			    "pesi.bytes_tree");
			lck_pesi_tree = Lck_CreateClass(&vsc_seg,
			    "pesi.pesi_tree");
			AN(lck_bytes_tree);
			AN(lck_pesi_tree);
		}

		VRT_AddVDP(ctx, &VDP_pesi);
		break;
	case VCL_EVENT_DISCARD:
		VRT_RemoveVDP(ctx, &VDP_pesi);

		AN(loadcnt);
		if (--loadcnt == 0)
			Lck_DestroyClass(&vsc_seg);
		break;
	case VCL_EVENT_WARM:
		if (warmcnt++ == 0) {
			AZ(mempool);
			mempool = mpl_init();
			AN(mempool);
		}
		break;
	case VCL_EVENT_COLD:
		AN(warmcnt);
		if (--warmcnt == 0) {
			AN(mempool);
			mpl_fini(&mempool);
			AZ(mempool);
		}
		break;
	default:
		WRONG("Illegal event enum");
	}
	return (0);
}

/* Functions */

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
